import os
import tempfile
from datetime import datetime
from io import StringIO
from unittest.mock import patch

import django
import pytz
from django.core import mail
from django.test import TestCase

from dbbackup import settings, utils
from dbbackup.tests.utils import (
    COMPRESSED_FILE,
    ENCRYPTED_FILE,
    add_private_gpg,
    add_public_gpg,
    callable_for_filename_template,
    clean_gpg_keys,
)


class Bytes_To_StrTest(TestCase):
    def test_get_gb(self):
        value = utils.bytes_to_str(byteVal=2**31)
        self.assertEqual(value, "2.0 GiB")

    def test_0_decimal(self):
        value = utils.bytes_to_str(byteVal=1.01, decimals=0)
        self.assertEqual(value, "1 B")

    def test_2_decimal(self):
        value = utils.bytes_to_str(byteVal=1.01, decimals=2)
        self.assertEqual(value, "1.01 B")


class Handle_SizeTest(TestCase):
    def test_func(self):
        filehandle = StringIO("Test string")
        value = utils.handle_size(filehandle=filehandle)
        self.assertEqual(value, "11.0 B")


class MailAdminsTest(TestCase):
    def test_func(self):
        subject = "foo subject"
        msg = "bar message"

        utils.mail_admins(subject, msg)
        self.assertEqual(len(mail.outbox), 1)

        sent_mail = mail.outbox[0]
        expected_subject = f"{settings.EMAIL_SUBJECT_PREFIX}{subject}"
        expected_to = settings.ADMINS[0][1]
        expected_from = settings.SERVER_EMAIL

        self.assertEqual(sent_mail.subject, expected_subject)
        self.assertEqual(sent_mail.body, msg)
        self.assertEqual(sent_mail.to[0], expected_to)
        self.assertEqual(sent_mail.from_email, expected_from)

    @patch("dbbackup.settings.ADMINS", None)
    def test_no_admin(self):
        subject = "foo subject"
        msg = "bar message"
        self.assertIsNone(utils.mail_admins(subject, msg))
        self.assertEqual(len(mail.outbox), 0)


class Email_Uncaught_ExceptionTest(TestCase):
    def test_success(self):
        def func():
            pass

        utils.email_uncaught_exception(func)
        self.assertEqual(len(mail.outbox), 0)

    @patch("dbbackup.settings.SEND_EMAIL", False)
    def test_raise_error_without_mail(self):
        def func():
            raise Exception("Foo")

        with self.assertRaises(Exception):
            utils.email_uncaught_exception(func)()
        self.assertEqual(len(mail.outbox), 0)

    @patch("dbbackup.settings.SEND_EMAIL", True)
    @patch("dbbackup.settings.FAILURE_RECIPIENTS", ["foo@bar"])
    def test_raise_with_mail(self):
        def func():
            raise Exception("Foo")

        with self.assertRaises(Exception):
            utils.email_uncaught_exception(func)()
        self.assertEqual(len(mail.outbox), 1)
        error_mail = mail.outbox[0]
        self.assertEqual(["foo@bar"], error_mail.to)
        self.assertIn('Exception("Foo")', error_mail.subject)
        if django.VERSION >= (1, 7):
            self.assertIn('Exception("Foo")', error_mail.body)


class Encrypt_FileTest(TestCase):
    def setUp(self):
        self.path = tempfile.mktemp()
        with open(self.path, "a") as fd:
            fd.write("foo")
        add_public_gpg()

    def tearDown(self):
        os.remove(self.path)
        clean_gpg_keys()

    def test_func(self, *args):
        with open(self.path) as fd:
            encrypted_file, filename = utils.encrypt_file(
                inputfile=fd, filename="foo.txt"
            )
        encrypted_file.seek(0)
        self.assertTrue(encrypted_file.read())


class Unencrypt_FileTest(TestCase):
    def setUp(self):
        add_private_gpg()

    def tearDown(self):
        clean_gpg_keys()

    @patch("dbbackup.utils.input", return_value=None)
    @patch("dbbackup.utils.getpass", return_value=None)
    def test_unencrypt(self, *args):
        inputfile = open(ENCRYPTED_FILE, "r+b")
        uncryptfile, filename = utils.unencrypt_file(inputfile, "foofile.gpg")
        uncryptfile.seek(0)
        self.assertEqual(b"foo\n", uncryptfile.read())


class Compress_FileTest(TestCase):
    def setUp(self):
        self.path = tempfile.mktemp()
        with open(self.path, "a+b") as fd:
            fd.write(b"foo")

    def tearDown(self):
        os.remove(self.path)

    def test_func(self, *args):
        with open(self.path) as fd:
            compressed_file, filename = utils.encrypt_file(
                inputfile=fd, filename="foo.txt"
            )


class Uncompress_FileTest(TestCase):
    def test_func(self):
        inputfile = open(COMPRESSED_FILE, "rb")
        fd, filename = utils.uncompress_file(inputfile, "foo.gz")
        fd.seek(0)
        self.assertEqual(fd.read(), b"foo\n")


class Create_Spooled_Temporary_FileTest(TestCase):
    def setUp(self):
        self.path = tempfile.mktemp()
        with open(self.path, "a") as fd:
            fd.write("foo")

    def tearDown(self):
        os.remove(self.path)

    def test_func(self, *args):
        utils.create_spooled_temporary_file(filepath=self.path)


class TimestampTest(TestCase):
    def test_naive_value(self):
        with self.settings(USE_TZ=False):
            timestamp = utils.timestamp(datetime(2015, 8, 15, 8, 15, 12, 0))
            self.assertEqual(timestamp, "2015-08-15-081512")

    def test_aware_value(self):
        with self.settings(USE_TZ=True) and self.settings(TIME_ZONE="Europe/Rome"):
            timestamp = utils.timestamp(
                datetime(2015, 8, 15, 8, 15, 12, 0, tzinfo=pytz.utc)
            )
            self.assertEqual(timestamp, "2015-08-15-101512")


class Datefmt_To_Regex(TestCase):
    def test_patterns(self):
        now = datetime.now()
        for datefmt, regex in utils.PATTERN_MATCHNG:
            date_string = datetime.strftime(now, datefmt)
            regex = utils.datefmt_to_regex(datefmt)
            match = regex.match(date_string)
            self.assertTrue(match)
            self.assertEqual(match.groups()[0], date_string)

    def test_complex_pattern(self):
        now = datetime.now()
        datefmt = "Foo%a_%A-%w-%d-%b-%B_%m_%y_%Y-%H-%I-%M_%S_%f_%j-%U-%W-Bar"
        date_string = datetime.strftime(now, datefmt)
        regex = utils.datefmt_to_regex(datefmt)
        self.assertTrue(regex.pattern.startswith("(Foo"))
        self.assertTrue(regex.pattern.endswith("Bar)"))
        match = regex.match(date_string)
        self.assertTrue(match)
        self.assertEqual(match.groups()[0], date_string)


class Filename_To_DatestringTest(TestCase):
    def test_func(self):
        now = datetime.now()
        datefmt = settings.DATE_FORMAT
        filename = f"{datetime.strftime(now, datefmt)}-foo.gz.gpg"
        datestring = utils.filename_to_datestring(filename, datefmt)
        self.assertIn(datestring, filename)

    def test_generated_filename(self):
        filename = utils.filename_generate("bak", "default")
        datestring = utils.filename_to_datestring(filename)
        self.assertIn(datestring, filename)


class Filename_To_DateTest(TestCase):
    def test_func(self):
        now = datetime.now()
        datefmt = settings.DATE_FORMAT
        filename = f"{datetime.strftime(now, datefmt)}-foo.gz.gpg"
        date = utils.filename_to_date(filename, datefmt)
        self.assertEqual(date.timetuple()[:5], now.timetuple()[:5])

    def test_generated_filename(self):
        filename = utils.filename_generate("bak", "default")
        utils.filename_to_date(filename)


@patch("dbbackup.settings.HOSTNAME", "test")
class Filename_GenerateTest(TestCase):
    @patch(
        "dbbackup.settings.FILENAME_TEMPLATE",
        "---{databasename}--{servername}-{datetime}.{extension}",
    )
    def test_func(self, *args):
        extension = "foo"
        generated_name = utils.filename_generate(extension)
        self.assertTrue("--" not in generated_name)
        self.assertFalse(generated_name.startswith("-"))

    def test_db(self, *args):
        extension = "foo"
        generated_name = utils.filename_generate(extension)
        self.assertTrue(generated_name.startswith(settings.HOSTNAME))
        self.assertTrue(generated_name.endswith(extension))

    def test_media(self, *args):
        extension = "foo"
        generated_name = utils.filename_generate(extension, content_type="media")
        self.assertTrue(generated_name.startswith(settings.HOSTNAME))
        self.assertTrue(generated_name.endswith(extension))

    @patch("django.utils.timezone.settings.USE_TZ", True)
    def test_tz_true(self):
        filename = utils.filename_generate("bak", "default")
        datestring = utils.filename_to_datestring(filename)
        self.assertIn(datestring, filename)

    @patch("dbbackup.settings.FILENAME_TEMPLATE", callable_for_filename_template)
    def test_template_is_callable(self, *args):
        extension = "foo"
        generated_name = utils.filename_generate(extension)
        self.assertTrue(generated_name.endswith("foo"))


class QuoteCommandArg(TestCase):
    def test_arg_with_space(self):
        assert utils.get_escaped_command_arg("foo bar") == "'foo bar'"



# DO NOT IMPORT THIS BEFORE django.configure() has been run!

import socket
import tempfile

from django.conf import settings

DATABASES = getattr(settings, "DBBACKUP_DATABASES", list(settings.DATABASES.keys()))

# Fake host
HOSTNAME = getattr(settings, "DBBACKUP_HOSTNAME", socket.gethostname())

# Directory to use for temporary files
TMP_DIR = getattr(settings, "DBBACKUP_TMP_DIR", tempfile.gettempdir())
TMP_FILE_MAX_SIZE = getattr(settings, "DBBACKUP_TMP_FILE_MAX_SIZE", 10 * 1024 * 1024)
TMP_FILE_READ_SIZE = getattr(settings, "DBBACKUP_TMP_FILE_READ_SIZE", 1024 * 1000)

# Number of old backup files to keep
CLEANUP_KEEP = getattr(settings, "DBBACKUP_CLEANUP_KEEP", 10)
CLEANUP_KEEP_MEDIA = getattr(settings, "DBBACKUP_CLEANUP_KEEP_MEDIA", CLEANUP_KEEP)
CLEANUP_KEEP_FILTER = getattr(settings, "DBBACKUP_CLEANUP_KEEP_FILTER", lambda x: False)

MEDIA_PATH = getattr(settings, "DBBACKUP_MEDIA_PATH", settings.MEDIA_ROOT)

DATE_FORMAT = getattr(settings, "DBBACKUP_DATE_FORMAT", "%Y-%m-%d-%H%M%S")
FILENAME_TEMPLATE = getattr(
    settings,
    "DBBACKUP_FILENAME_TEMPLATE",
    "{databasename}-{servername}-{datetime}.{extension}",
)
MEDIA_FILENAME_TEMPLATE = getattr(
    settings, "DBBACKUP_MEDIA_FILENAME_TEMPLATE", "{servername}-{datetime}.{extension}"
)

GPG_ALWAYS_TRUST = getattr(settings, "DBBACKUP_GPG_ALWAYS_TRUST", False)
GPG_RECIPIENT = GPG_ALWAYS_TRUST = getattr(settings, "DBBACKUP_GPG_RECIPIENT", None)

STORAGE = getattr(settings, "DBBACKUP_STORAGE", None)
STORAGE_OPTIONS = getattr(settings, "DBBACKUP_STORAGE_OPTIONS", {})
# https://docs.djangoproject.com/en/5.1/ref/settings/#std-setting-STORAGES
STORAGES_DBBACKUP_ALIAS = "dbbackup"
DJANGO_STORAGES = getattr(settings, "STORAGES", {})
django_dbbackup_storage = DJANGO_STORAGES.get(STORAGES_DBBACKUP_ALIAS, {})

if not STORAGE:
    STORAGE = (
        django_dbbackup_storage.get("BACKEND")
        or "django.core.files.storage.FileSystemStorage"
    )
if not STORAGE_OPTIONS:
    STORAGE_OPTIONS = django_dbbackup_storage.get("OPTIONS") or STORAGE_OPTIONS

CONNECTORS = getattr(settings, "DBBACKUP_CONNECTORS", {})

CUSTOM_CONNECTOR_MAPPING = getattr(settings, "DBBACKUP_CONNECTOR_MAPPING", {})

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

# Mail
SEND_EMAIL = getattr(settings, "DBBACKUP_SEND_EMAIL", True)
SERVER_EMAIL = getattr(settings, "DBBACKUP_SERVER_EMAIL", settings.SERVER_EMAIL)
FAILURE_RECIPIENTS = getattr(settings, "DBBACKUP_FAILURE_RECIPIENTS", None)
if FAILURE_RECIPIENTS is None:
    ADMINS = getattr(settings, "DBBACKUP_ADMIN", settings.ADMINS)
else:
    ADMINS = FAILURE_RECIPIENTS
EMAIL_SUBJECT_PREFIX = getattr(settings, "DBBACKUP_EMAIL_SUBJECT_PREFIX", "[dbbackup] ")



import logging
from typing import List, Optional
from urllib.parse import quote

from .base import BaseCommandDBConnector

logger = logging.getLogger("dbbackup.command")


def create_postgres_uri(self):
    host = self.settings.get("HOST") or "localhost"
    dbname = self.settings.get("NAME") or ""
    user = quote(self.settings.get("USER") or "")
    password = self.settings.get("PASSWORD") or ""
    password = f":{quote(password)}" if password else ""
    if not user:
        password = ""
    else:
        host = "@" + host

    port = ":{}".format(self.settings.get("PORT")) if self.settings.get("PORT") else ""
    dbname = f"--dbname=postgresql://{user}{password}{host}{port}/{dbname}"
    return dbname


class PgDumpConnector(BaseCommandDBConnector):
    """
    PostgreSQL connector, it uses pg_dump`` to create an SQL text file
    and ``psql`` for restore it.
    """

    extension = "psql"
    dump_cmd = "pg_dump"
    restore_cmd = "psql"
    single_transaction = True
    drop = True
    schemas: Optional[List[str]] = []

    def _create_dump(self):
        cmd = f"{self.dump_cmd} "
        cmd = cmd + create_postgres_uri(self)

        for table in self.exclude:
            cmd += f" --exclude-table-data={table}"

        if self.drop:
            cmd += " --clean"

        if self.schemas:
            # First schema is not prefixed with -n
            # when using join function so add it manually.
            cmd += " -n " + " -n ".join(self.schemas)

        cmd = f"{self.dump_prefix} {cmd} {self.dump_suffix}"
        stdout, stderr = self.run_command(cmd, env=self.dump_env)
        return stdout

    def _restore_dump(self, dump):
        cmd = f"{self.restore_cmd} "
        cmd = cmd + create_postgres_uri(self)

        # without this, psql terminates with an exit value of 0 regardless of errors
        cmd += " --set ON_ERROR_STOP=on"

        if self.schemas:
            cmd += " -n " + " -n ".join(self.schemas)

        if self.single_transaction:
            cmd += " --single-transaction"

        cmd += " {}".format(self.settings["NAME"])
        cmd = f"{self.restore_prefix} {cmd} {self.restore_suffix}"
        stdout, stderr = self.run_command(cmd, stdin=dump, env=self.restore_env)
        return stdout, stderr


class PgDumpGisConnector(PgDumpConnector):
    """
    PostgreGIS connector, same than :class:`PgDumpGisConnector` but enable
    postgis if not made.
    """

    psql_cmd = "psql"

    def _enable_postgis(self):
        cmd = f'{self.psql_cmd} -c "CREATE EXTENSION IF NOT EXISTS postgis;"'
        cmd += " --username={}".format(self.settings["ADMIN_USER"])
        cmd += " --no-password"

        if self.settings.get("HOST"):
            cmd += " --host={}".format(self.settings["HOST"])

        if self.settings.get("PORT"):
            cmd += " --port={}".format(self.settings["PORT"])

        return self.run_command(cmd)

    def _restore_dump(self, dump):
        if self.settings.get("ADMIN_USER"):
            self._enable_postgis()
        return super()._restore_dump(dump)


class PgDumpBinaryConnector(PgDumpConnector):
    """
    PostgreSQL connector, it uses pg_dump`` to create an SQL text file
    and ``pg_restore`` for restore it.
    """

    extension = "psql.bin"
    dump_cmd = "pg_dump"
    restore_cmd = "pg_restore"
    single_transaction = True
    drop = True

    def _create_dump(self):
        cmd = f"{self.dump_cmd} "
        cmd = cmd + create_postgres_uri(self)

        cmd += " --format=custom"
        for table in self.exclude:
            cmd += f" --exclude-table-data={table}"

        if self.schemas:
            cmd += " -n " + " -n ".join(self.schemas)

        cmd = f"{self.dump_prefix} {cmd} {self.dump_suffix}"
        stdout, _ = self.run_command(cmd, env=self.dump_env)
        return stdout

    def _restore_dump(self, dump):
        dbname = create_postgres_uri(self)
        cmd = f"{self.restore_cmd} {dbname}"

        if self.single_transaction:
            cmd += " --single-transaction"

        if self.drop:
            cmd += " --clean"

        if self.schemas:
            cmd += " -n " + " -n ".join(self.schemas)

        cmd = f"{self.restore_prefix} {cmd} {self.restore_suffix}"
        stdout, stderr = self.run_command(cmd, stdin=dump, env=self.restore_env)
        return stdout, stderr



from unittest.mock import patch

from django.test import TestCase

try:
    from dbbackup import checks
    from dbbackup.apps import DbbackupConfig
except ImportError:
    checks = None


def test_func(*args, **kwargs):
    return "foo"


class ChecksTest(TestCase):
    def setUp(self):
        if checks is None:
            self.skipTest("Test framework has been released in Django 1.7")

    def test_check(self):
        self.assertFalse(checks.check_settings(DbbackupConfig))

    @patch("dbbackup.checks.settings.HOSTNAME", "")
    def test_hostname_invalid(self):
        expected_errors = [checks.W001]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)

    @patch("dbbackup.checks.settings.STORAGE", "")
    def test_hostname_storage(self):
        expected_errors = [checks.W002]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)

    @patch("dbbackup.checks.settings.FILENAME_TEMPLATE", test_func)
    def test_filename_template_is_callable(self):
        self.assertFalse(checks.check_settings(DbbackupConfig))

    @patch("dbbackup.checks.settings.FILENAME_TEMPLATE", "{datetime}.bak")
    def test_filename_template_is_string(self):
        self.assertFalse(checks.check_settings(DbbackupConfig))

    @patch("dbbackup.checks.settings.FILENAME_TEMPLATE", "foo.bak")
    def test_filename_template_no_date(self):
        expected_errors = [checks.W003]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)

    @patch("dbbackup.checks.settings.MEDIA_FILENAME_TEMPLATE", test_func)
    def test_media_filename_template_is_callable(self):
        self.assertFalse(checks.check_settings(DbbackupConfig))

    @patch("dbbackup.checks.settings.MEDIA_FILENAME_TEMPLATE", "{datetime}.bak")
    def test_media_filename_template_is_string(self):
        self.assertFalse(checks.check_settings(DbbackupConfig))

    @patch("dbbackup.checks.settings.MEDIA_FILENAME_TEMPLATE", "foo.bak")
    def test_media_filename_template_no_date(self):
        expected_errors = [checks.W004]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)

    @patch("dbbackup.checks.settings.DATE_FORMAT", "foo@net.pt")
    def test_date_format_warning(self):
        expected_errors = [checks.W005]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)

    @patch("dbbackup.checks.settings.FAILURE_RECIPIENTS", "foo@net.pt")
    def test_Failure_recipients_warning(self):
        expected_errors = [checks.W006]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)

    @patch("dbbackup.checks.settings.FILENAME_TEMPLATE", "foo/bar-{datetime}.ext")
    def test_db_filename_template_with_slash(self):
        expected_errors = [checks.W007]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)

    @patch("dbbackup.checks.settings.FILENAME_TEMPLATE", lambda _: "foo/bar")
    def test_db_filename_template_callable_with_slash(self):
        expected_errors = [checks.W007]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)

    @patch("dbbackup.checks.settings.MEDIA_FILENAME_TEMPLATE", "foo/bar-{datetime}.ext")
    def test_media_filename_template_with_slash(self):
        expected_errors = [checks.W008]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)

    @patch("dbbackup.checks.settings.MEDIA_FILENAME_TEMPLATE", lambda _: "foo/bar")
    def test_media_filename_template_callable_with_slash(self):
        expected_errors = [checks.W008]
        errors = checks.check_settings(DbbackupConfig)
        self.assertEqual(expected_errors, errors)
